/****************************************************************/
/**@file PCD8006.c:
 * Description: API layer for TEDIA PCD-8006 card.              *
 * Dependency: Windows 32, Windows 64 or Linux                  *
 *                Copyright 2019-2020 Jaroslav Fojtik           *
 ****************************************************************/
#if defined(_WIN32) || defined(_WIN64)
  #include <windows.h>
#else
  #include <sys/io.h>
#endif
#include <malloc.h>
#include <math.h>

#include "hudaqlib.h"
#include "hudaq_internal.h"


#define DI_CHANNELS   6		///< Amount of Digital Outputs
#define DO_CHANNELS   6		///< Amount of Digital Inputs
#define DA_CHANNELS   0         ///< Amount of Analog Outputs
#define AD_CHANNELS   0         ///< Amount of Analog Inputs
#define ENC_CHANNELS  0         ///< Amount of Encoders
#define CTR_CHANNELS  1         ///< Amount of counters
#define STEP_CHANNELS 0         ///< Amount of steppers


#define MASTERFREQUENCY 1000        ///< Internal oscillator frequency is 1kHz


/** Cache of PCD-7004 state */
typedef struct
{
  UserDataHeader Hdr;                   ///< General device setup

  unsigned __int8 DIOReg[6];		///< Port configuration register.
  unsigned __int8 DIOCfgReg;		///< Port direction settings.
  unsigned __int8 IRQCfgReg;
  unsigned __int8 INTEnReg;
  unsigned __int8 TimerReg;
} PCD_8006_Private;


/** Symbolic aliases for available registers inside BADR4 read. */
typedef enum
{
  DIOReg0    = 0,
  DIOReg1    = 1,	//4
  DIOReg2    = 2,	//8,
  DIOReg3    = 3,	//0xC,
  DIOReg4    = 4,	//0x10,
  DIOReg5    = 5,	//0x14,

  DIOCfgReg  = 0x80/4,

  IRQCfgReg  = 0x200/4,		// WR only
  IRQStatusReg = 0x200/4,	// RD only
  IRQClrReg =  0x204/4,		// WR only
  
  TimerReg =  0x208/4,		// RW
  INTEnReg =  0x20C/4,

} MEM4IO;

static __int8 Ports[DI_CHANNELS] = {DIOReg0, DIOReg1, DIOReg2, DIOReg3, DIOReg4, DIOReg5};


/** write to memory mapped device byte wise, bytes are stored on every 4th position. */
static __inline void StoreByte(size_t Ptr, int Offset, unsigned __int8 value)
{
  ((volatile unsigned __int8 *)(Ptr))[4*Offset] = value;
}

/** write to memory mapped device byte wise through cache. */
static __inline void StoreCachedByte(size_t Ptr, int Offset, unsigned __int8 value, unsigned __int8 *CachedValue)
{
  if (*CachedValue != value)
    {
    *CachedValue = value;
    StoreByte(Ptr,Offset,value);
    }
}


/** This inline is used for reading tightly packed OX9162 local configuration registers. */
static __inline unsigned __int8 GetBytePlain(size_t Ptr, int Offset)
{
  return ((volatile unsigned __int8 *)(Ptr))[4*Offset];
}


/****************************************************************
 *                                                              *
 *                 DIGITAL INPUTS & OUTPUTS                     *
 *                                                              *
 ****************************************************************/


/** Get data from digital input. */
static int PCD8006DIRead(const DeviceRecord *DevRecord, unsigned channel)
{
  if (channel>=DI_CHANNELS) return HUDAQBADARG;
  return GetBytePlain(DevRecord->DrvRes.Resources.MemResources[0].Base,Ports[channel]);
}


/** Write data to digital output. */
static HUDAQSTATUS PCD8006DOWrite(const DeviceRecord *DevRecord, unsigned channel, unsigned value)
{
  if (channel>=DO_CHANNELS) return HUDAQBADARG;
  StoreCachedByte(DevRecord->DrvRes.Resources.MemResources[0].Base, Ports[channel], value,
                  &((PCD_8006_Private *)(DevRecord->DrvRes.DriverData))->DIOReg[channel] );
  return HUDAQSUCCESS;
}


static HUDAQSTATUS PCD8006DIReadMultiple(const DeviceRecord *DevRecord, unsigned number, const unsigned *channels, unsigned *values)
{
unsigned FlagX = 0;
unsigned i;
size_t Base;
unsigned __int8 Cache[DI_CHANNELS+1];  //allocate more bytes for padding ind

  if(number==0 || channels==NULL || values==NULL) return HUDAQBADARG;
  if(DevRecord->pCT->HudaqDIRead==NULL) return HUDAQFAILURE;

  Base = DevRecord->DrvRes.Resources.MemResources[0].Base;

	/* Schedule what to read. */
  for(i=0; i<number; i++)
  {
    if(channels[i]<DI_CHANNELS)
      {
      FlagX |= 1<<(channels[i]);
      }
    else
      {
      values[i] = 0;
      FlagX |= 0x8000;
      }
    }
    
    /* Read isolated DIs */
  for(i=0; i<DI_CHANNELS; i++)
    {
    if(FlagX & (1<<i))
      {
      Cache[i] = GetBytePlain(Base,Ports[i]);
      }
    }  
    
	/* Final store of numbers read. */
  for(i=0; i<number; i++)
  {
    if(channels[i]<DI_CHANNELS)
      {
      values[i] = Cache[channels[i]];
      }
  }      
  
  if(FlagX == 0x8000) return HUDAQFAILURE;	//no valid value returned
  if(FlagX & 0x8000) return HUDAQPARTIAL;	//several values are valid and several are not
  return HUDAQSUCCESS;			//all values are valid (or no channel asked)
}


static HUDAQSTATUS PCD8006DOWriteMultiple(const DeviceRecord *DevRecord, unsigned number, const unsigned *channels, const unsigned *values)
{
unsigned FlagX = 0;
unsigned i;
size_t Base;
unsigned __int8 *Cache;

  if (number==0 || channels==NULL || values==NULL) return HUDAQBADARG;
  if(DevRecord->pCT->HudaqDIRead==NULL) return HUDAQFAILURE;

  Base = DevRecord->DrvRes.Resources.MemResources[0].Base;
  Cache = ((PCD_8006_Private *)(DevRecord->DrvRes.DriverData))->DIOReg;

    /* schedule what to write */
  for(i=0; i<number; i++)
  {
    if (channels[i]<DI_CHANNELS)
      {
      if(Cache[i] != values[i])
        {
        FlagX |= 1<<(channels[i]);
        Cache[i] = values[i];
        }
      }
    else
      {
      FlagX |= 0x8000;
      }
    }

    /* Final store of DOs */
  for(i=0; i<DI_CHANNELS; i++)
    {
    if(FlagX & (1<<i))
      {
      StoreByte(Base, Ports[i], Cache[i]);
      }
    }  
      
  if(FlagX == 0x8000) return HUDAQFAILURE;	//no valid value returned
  if(FlagX & 0x8000) return HUDAQPARTIAL;	//several values are valid and several are not
  return HUDAQSUCCESS;			//all values are valid (or no channel asked)
}


/** Write one bit to digital output. */
static void PCD8006DOWriteBit(const DeviceRecord *DevRecord, unsigned channel, unsigned bit, int value)
{
__int8 DoutByte;

  if (channel>=DO_CHANNELS) return;
  if (bit>=8) return;		/* There are only 8 bits available in PCD8006. */

  DoutByte = (((PCD_8006_Private *)(DevRecord->DrvRes.DriverData))->DIOReg[channel]);
  if (value) DoutByte |= (1<<bit);
       else DoutByte &= ~(1<<bit);

  StoreCachedByte(DevRecord->DrvRes.Resources.MemResources[0].Base, Ports[channel], DoutByte,
                  & ((PCD_8006_Private *)(DevRecord->DrvRes.DriverData))->DIOReg[channel]);
}


static void PCD8006DOWriteMultipleBits(const DeviceRecord *DevRecord, unsigned channel, unsigned mask, unsigned value)
{
__int8 DoutByte;

  if (channel>=DO_CHANNELS) return;

  DoutByte = (((PCD_8006_Private *)(DevRecord->DrvRes.DriverData))->DIOReg[channel]);
  DoutByte = (DoutByte & ~mask) | (value & mask);

  StoreCachedByte(DevRecord->DrvRes.Resources.MemResources[0].Base, Ports[channel], DoutByte,
                  & ((PCD_8006_Private *)(DevRecord->DrvRes.DriverData))->DIOReg[channel] );
}


static double PCD8006DIOGetParameter(const DeviceRecord *DevRecord, unsigned channel, HudaqParameter param)
{
unsigned __int8 chval;

  switch((int)param)
    {
    case HudaqDINUMBITS:
           return (channel<DI_CHANNELS) ? 8 : WRONG_VALUE;
    case HudaqDONUMBITS:
           return (channel<DO_CHANNELS) ? 8 : WRONG_VALUE;
    case HudaqDINUMCHANNELS:
	   return DI_CHANNELS;
    case HudaqDONUMCHANNELS:
	   return DO_CHANNELS;
    case HudaqDOMODE:
           if(channel>=DO_CHANNELS) return WRONG_VALUE;
           chval = ((PCD_8006_Private *)(DevRecord->DrvRes.DriverData))->DIOCfgReg;
           return (chval>>channel) & 1;
    }

  return WRONG_VALUE;
}


static HUDAQSTATUS PCD8006DIOSetParameter(const DeviceRecord *DevRecord, unsigned channel, HudaqParameter param, int value)
{
unsigned __int8 chval;

  switch((int)param)
    {
    case HudaqDINUMBITS:
           if(channel>=DI_CHANNELS) return WRONG_VALUE;
           return (value==8) ? 0 : WRONG_VALUE;
    case HudaqDONUMBITS:
           if(channel>=DI_CHANNELS) return WRONG_VALUE;
           return (value==8) ? 0 : WRONG_VALUE;
    case HudaqDINUMCHANNELS:
	   return DI_CHANNELS;
    case HudaqDONUMCHANNELS:
	   return DO_CHANNELS;
    case HudaqDOMODE:
          if(channel>=DO_CHANNELS) return WRONG_VALUE;
          chval = ((PCD_8006_Private *)(DevRecord->DrvRes.DriverData))->DIOCfgReg;
	  //printf("chval =  %d\n",chval);          
     
          switch(channel)
          {
            case 0: if(value==1) chval &= ~0x1;
        	    if(value==0) chval |= 0x1;
        	    break;
            case 1: if(value==1) chval &= ~0x02;
        	    if(value==0) chval |= 0x02;
        	    break;
            case 2: if(value==1) chval &= ~0x04;
        	    if(value==0) chval |= 0x04;
		    break;
            case 3: if(value==1) chval &= ~0x08;
        	    if(value==0) chval |= 0x08;
        	    break;
            case 4: if(value==1) chval &= ~0x10;
        	    if(value==0) chval |= 0x10;
        	    break;
            case 5: if(value==1) chval &= ~0x20;
        	    if(value==0) chval |= 0x20;
        	    break;
          }
          StoreCachedByte(DevRecord->DrvRes.Resources.MemResources[0].Base, DIOCfgReg, chval,
                  & ((PCD_8006_Private *)(DevRecord->DrvRes.DriverData))->DIOCfgReg );
          return HUDAQSUCCESS;
    }

  return WRONG_VALUE;
}


/****************************************************************
 *                                                              *
 *                COUNTERS PWM + Count + Step                   *
 *                                                              *
 ****************************************************************/

static double PCD8006CtrGetParameter(unsigned channel, HudaqParameter param)
{
  if (channel>=CTR_CHANNELS) return WRONG_VALUE;

  switch(param)
    {
    case HudaqPwmNUMCHANNELS:
    case HudaqCtrNUMCHANNELS:
           return CTR_CHANNELS;
    }

  return WRONG_VALUE;
}


static HUDAQSTATUS PCD8006PWMWrite(const DeviceRecord *DevRecord, unsigned channel, double frequency, double dutycycle)
{
double T;

  if (channel>=CTR_CHANNELS) return HUDAQBADARG;

  if (frequency<=0) return HUDAQBADARG; //Allow duty cycle 0 and 1 for f=0.

  T = MASTERFREQUENCY/frequency;
  if(T<1) 
  {
    StoreCachedByte(DevRecord->DrvRes.Resources.MemResources[0].Base, TimerReg, 1,
                  & ((PCD_8006_Private *)(DevRecord->DrvRes.DriverData))->TimerReg);
    return HUDAQPARTIAL;
  }
  if(T>255)
  {
    StoreCachedByte(DevRecord->DrvRes.Resources.MemResources[0].Base, TimerReg, 255,
                  & ((PCD_8006_Private *)(DevRecord->DrvRes.DriverData))->TimerReg);
    return HUDAQPARTIAL;
  }

  StoreCachedByte(DevRecord->DrvRes.Resources.MemResources[0].Base, TimerReg, 
		  (unsigned __int8)T,
                  &((PCD_8006_Private *)(DevRecord->DrvRes.DriverData))->TimerReg);

  return HUDAQSUCCESS;
}


static HUDAQSTATUS PCD8006CtrReset(const DeviceRecord *DevRecord, unsigned channel)
{
  if (channel>=CTR_CHANNELS) return HUDAQBADARG;
  StoreCachedByte(DevRecord->DrvRes.Resources.MemResources[0].Base, TimerReg, 
		  0, &((PCD_8006_Private *)(DevRecord->DrvRes.DriverData))->TimerReg);
  return HUDAQSUCCESS;
}


static int PCD8006CtrRead(const DeviceRecord *DevRecord, unsigned channel)
{
  if (channel>=CTR_CHANNELS) return 0;
  return GetBytePlain(DevRecord->DrvRes.Resources.MemResources[0].Base, TimerReg);
}





/********************************************************************/

/* Initialization procedure for PCD-8006 card. */
static int InitPCD8006(DeviceRecord *DevRecord, int IniOptions)
{
PCD_8006_Private *Cache;
int i;
size_t Base;

  if(DevRecord==NULL) return -1;

  if(DevRecord->DrvRes.Resources.NumMemResources<2) return -2;  //insufficient amount of resources
  Cache = DevRecord->DrvRes.DriverData;
  if(Cache==NULL) return -3;
  if(DevRecord->DrvRes.DriverDataSize<sizeof(PCD_8006_Private)) return -4;

  DevRecord->pCT = &CTPCD8006;

  if((IniOptions & HudaqOpenNOINIT)==0)
  {                           //initialize hardware only if no other application is running
    Base = DevRecord->DrvRes.Resources.MemResources[0].Base;

	/* disable interrupts */
    Cache->INTEnReg = 0;
    StoreByte(Base, INTEnReg, 0);

      /* Initialization of DI and DO. */
    for(i=0; i<6; i++)
    {
      Cache->DIOReg[i] = 0;
      StoreByte(Base, Ports[i], 0);
    }

    Cache->DIOCfgReg = 0;
    StoreByte(Base, DIOCfgReg, 0);

    Cache->IRQCfgReg = 0;
    StoreByte(Base, IRQCfgReg, 0);

    Cache->TimerReg = 0;
    StoreByte(Base, TimerReg, 0);	/* Stop timer. */
  }

return 1; //success
}


/** Internal cleanup procedure for PCD8006 */
static void DonePCD8006(DeviceRecord *DevRecord)
{
  if (DevRecord==NULL) return;

  DevRecord->pCT = &CtDummy;
}



/****************************************************************
 *                                                              *
 *                    GET/SET PARAMETERS                        *
 *                                                              *
 ****************************************************************/

static HUDAQSTATUS PCD8006IRQSetParameter(const DeviceRecord *DevRecord, unsigned channel, HudaqParameter param, int value)
{
size_t Base;

  Base = DevRecord->DrvRes.Resources.MemResources[0].Base;

  switch((int)param)
    {
    case HudaqIRQ+0:
	{
	 if(value)
	 {
	   StoreCachedByte(Base, IRQCfgReg, 1, &((PCD_8006_Private *)DevRecord->DrvRes.DriverData)->IRQCfgReg);
	   StoreByte(Base, IRQClrReg, 0xFF);
	   StoreCachedByte(Base, INTEnReg, 0x80, &((PCD_8006_Private *)DevRecord->DrvRes.DriverData)->INTEnReg);
         }
         else
	 { 
	   StoreCachedByte(DevRecord->DrvRes.Resources.MemResources[0].Base, INTEnReg, 0,
			   &((PCD_8006_Private *)DevRecord->DrvRes.DriverData)->INTEnReg);
	   StoreCachedByte(DevRecord->DrvRes.Resources.MemResources[0].Base, IRQCfgReg, 0,
			   &((PCD_8006_Private *)DevRecord->DrvRes.DriverData)->IRQCfgReg);
         }
        return HUDAQSUCCESS;
        }
    case HudaqIRQ+1:
           return HUDAQSUCCESS;
    }

  return WRONG_VALUE;
}


static double PCD8006IRQGetParameter(const DeviceRecord *DevRecord, unsigned channel, HudaqParameter param)
{
  switch((int)param)
    {
    case HudaqIRQ+0: 
	   return ((PCD_8006_Private *)DevRecord->DrvRes.DriverData)->Hdr.IRQcounter;
    case HudaqIRQ+1:
           return GetBytePlain(DevRecord->DrvRes.Resources.MemResources[0].Base,IRQStatusReg);
    }
  return WRONG_VALUE;
}


static HUDAQSTATUS PCD8006SetParameter(const DeviceRecord *DevRecord, unsigned channel, HudaqParameter param, double value)
{
switch(param & HudaqSubsystemMASK)
  {
  case HudaqDI:
  case HudaqDO:  return PCD8006DIOSetParameter(DevRecord,channel,param,(int)value);
//  case HudaqAI:  
//  case HudaqAO:  
//  case HudaqEnc: 
//  case HudaqPWM:
//  case HudaqCtr: 
//  case HudaqStep:
    case HudaqIRQ: return PCD8006IRQSetParameter(DevRecord,channel,param,(int)value);
  }
return HUDAQNOTSUPPORTED;
}

static double PCD8006GetParameter(const DeviceRecord *DevRecord, unsigned channel, HudaqParameter param)
{
switch(param & HudaqSubsystemMASK)
  {
  case HudaqDI:
  case HudaqDO:  return PCD8006DIOGetParameter(DevRecord,channel,param);
//case HudaqAI:  
//case HudaqAO:  
//case HudaqEnc:
//case HudaqPWM:
  case HudaqCtr: return PCD8006CtrGetParameter(channel,param);
//case HudaqStep:
  case HudaqIRQ: return PCD8006IRQGetParameter(DevRecord,channel,param);
  }
return WRONG_VALUE;
}


const CallTable CTPCD8006 =
    {
    "PCD-8006", 0x1760, 0x0800,
    InitPCD8006,
    DonePCD8006,

    PCD8006SetParameter,
    PCD8006GetParameter,
    NULL,

        // INITIALIZE DI callers
    PCD8006DIRead,
    GenericDIReadBit,           //Generic implementation
    PCD8006DIReadMultiple,
        // INITIALIZE DO callers
    PCD8006DOWrite,
    PCD8006DOWriteBit,
    PCD8006DOWriteMultipleBits,
    PCD8006DOWriteMultiple,
        // INITIALIZE AI callers
    NULL,
    NULL,
        // INITIALIZE AO callers
    NULL,
    NULL,
        // INITIALIZE Enc callers
    NULL,                       // Not available
    NULL,
        // INITIALIZE Ctr callers
    PCD8006CtrRead,
    PCD8006CtrReset,
        // INITIALIZE PWM callers
    PCD8006PWMWrite,            
    NULL,			// Not available
    NULL,
        // INITIALIZE Step callers
    NULL,                       // Not available
    };
    